/* element.c - Message Element Format implementation
 *
 * Copyright (C) Thomas Vander Stichele
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* the Message Element Format structure is as follows:
 * MEC		Message Element Code		1 byte
 * [DSN]	Data Set Number			0 or 1 byte
 *						0 means current
 *						254 means all but current
 *						255 means all
 * [PSN]	Program Service Number		0 or 1 byte
 *						0 means main service
 * [MEL]	Message Element Data Length	0 or 1 byte
 * [MED]	Message Element Data		0 .. 254 bytes
 */

#include <glib.h>
#include "element.h"

/* public functions */

/* create a new uecp element and initialize it, and return it */
UecpElement *
uecp_element_new ()
{
	UecpElement *element;

	element = g_malloc (sizeof (UecpElement));
	element->message_element_data = NULL; /* to avoid crashing g_free */
	uecp_element_reset (element);

	return element;
}

/* reset a element to a zero status */
void
uecp_element_reset (UecpElement *element)
{
	element->message_element_code = 0;
	element->data_set_number = -1;
	element->program_service_number = -1;
	element->message_element_length = -1;
	element->message_element_data_length = 0;
	if (element->message_element_data != NULL)
	{
		g_free (element->message_element_data);
		element->message_element_data = NULL;
	}
}

/* dispose of all resources held by element */
void
uecp_element_dispose (UecpElement *element)
{
	uecp_element_reset (element);
	g_free (element);
}


/* get the length of the given element
 * used both for allocation and housekeeping of length of blocks
 * (since they have no terminator */

unsigned int
uecp_element_get_length (UecpElement *element)
{
	int length = 0;

	/* MEL */
	if (element->message_element_length > 0)
		++length;
	/* MED */
	if (element->message_element_data_length >= 0)
		length += element->message_element_data_length;
	/* MEC */
	length++;
	/* DSN */
	if (element->data_set_number >= 0) ++length;
	/* PSN */
	if (element->program_service_number >= 0) ++length;

	return length;
}

/* convert our UecpElement structure to a sequence of bytes that represents
 * one element in the UECP format
 * make sure that the block to render to has enough place to contain it !
 */
void
uecp_element_render_to_block (UecpElement *element, unsigned char *block)
{
	int length;
	unsigned char *ptr;

	ptr = block;
	*ptr = element->message_element_code;
	ptr++;
	if (element->data_set_number >= 0)
	{
		*ptr = element->data_set_number;
		++ptr;
	}
	if (element->program_service_number >= 0)
	{
		*ptr = element->program_service_number;
		++ptr;
	}
	/* only render MEL if it's specified */
	if (element->message_element_length >= 0)
	{
		*ptr = element->message_element_length;
		++ptr;
	}
	/* regardless of MEL, it can still have a size; see PS */
	if (element->message_element_data_length >= 0)
	{
		memcpy (ptr, element->message_element_data,
			element->message_element_data_length);
	}
}

/* set all of the properties of an element at once
 * -1 for dsn, psn or mel means that it should not be taken into account */
gboolean
uecp_element_set (UecpElement *element, unsigned char mec,
		  int dsn, int psn, int mel, int medl, unsigned char *med)
{
	/* boundary checks */
	if (mec < 1)
	{
		g_print ("uecp: error: counter < 1\n");
		return FALSE;
	}

	element->message_element_code = mec;
	if (dsn >= 0) element->data_set_number = dsn;
	if (psn >= 0) element->program_service_number = psn;
	if (mel > 0) element->message_element_length = mel;
	if (medl > 0) element->message_element_data_length = medl;
	if (medl > 0) element->message_element_data = med;
	return TRUE;
}
