
/* frame.h - General Frame Format implementation
 *
 * Copyright (C) Thomas Vander Stichele
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * the General Frame is the output block for UECP commands.
 * structure is as follows:
 * STA		Start byte		1 byte (0xfe)
 * ADD		Address			2 bytes (10 bit site, 6 bit encoder)
 *					value of 0 means "ALL"
 * SQC		Sequence Counter	1 byte
 *					value of 0 means "unused"
 * MFL		Message Field Length	1 byte
 * [MSG]	Message			0 .. 255 bytes (optional)
 * CRC		Cyclic Redundancy Code	2 bytes
 * STP		Stop			1 byte (0xff)
 */

#include <glib.h>
#include "element.h"
#include "frame.h"

/* maximum field length */
#define	UECP_MAX_FIELD_LENGTH	255

/* external functions */
/* from crc.c */
extern guint16 ccitt_crc (const unsigned char *ptr, register int len);

/* private functions */

/*
 * calculate the length this block would have after stuffing
 * ie, return length + (number of bytes equal to FD, FE or FF)
 */
static unsigned int
uecp_byte_stuff_get_length (unsigned char *block, unsigned int length)
{
	unsigned int value = length;
	int i;

	for (i = 0; i < length; ++i)
	{
		if (block[i] >= 0xfd) ++value;
	}
	return value;
}

/*
 * stuff the given block with the given length and output to the
 * pre-allocated buffer given
 * byte stuffing does FD -> FD 00, FE -> FD 01, FF -> FD 02
 */
static void
uecp_byte_stuff_to_block (unsigned char *block, unsigned int length,
		          unsigned char *dest)
{
	unsigned int stuffed_length;
	unsigned char *p, *q;

	stuffed_length = uecp_byte_stuff_get_length (block, length);
	q = dest;
	for (p = block; p < block + length; ++p)
	{
		if (*p >= 0xfd)
		{
			*q = 0xfd;
			++q;
			*q = *p - 0xfd;
		}
		else
		{
			*q = *p;
		}
		++q;
	}
	if (stuffed_length != q - dest)
		g_print ("uecp: ERROR: stuffed length %d, actual one %d\n",
		         stuffed_length, q - dest);
}

/* public functions */

/* create a new uecp frame and initialize it, and return it */
UecpFrame *
uecp_frame_new ()
{
	UecpFrame *frame;

	frame = g_malloc (sizeof (UecpFrame));
	frame->field = NULL;		/* to avoid a crashing free */
	uecp_frame_reset (frame);

	return frame;
}

/* reset a frame to a zero status */
void
uecp_frame_reset (UecpFrame *frame)
{
	frame->site_address = 0;
	frame->encoder_address = 0;
	frame->sequence_counter = 0;
	frame->field_length = 0;
	if (frame->field != NULL)
	{
		/* FIXME:
		 * free the list one by one
		uecp_field_dispose (frame->field);
		*/
		frame->field = NULL;
	}
}

/* dispose of all resources held by frame */
void
uecp_frame_dispose (UecpFrame *frame)
{
	uecp_frame_reset (frame);
	g_free (frame);
}


/* convert our UecpFrame structure to a sequence of bytes that represents
 * one frame in the UECP format
 * returns a newly allocated block which the caller must free afterwards
 */
unsigned char *
uecp_frame_render (UecpFrame *frame)
{
	/* first, assemble the byte sequence to crc
	 * this includes 2 byte address, 1 byte SQC, 1 byte MFL, and message
	 * */
	unsigned int crc;
	int crc_length;
	unsigned char *block;
	int block_length;
	unsigned char *stuffed;
	int stuffed_length;

	crc_length = 2 + 1 + 1 + frame->field_length;
	block_length = crc_length + 2; /* length of block to byte stuff */
	block = (unsigned char *) g_malloc (block_length);
	/* ADD field: 10 bit site address + 6 bit encoder address */
	block[0] = frame->site_address / 4;
	block[1] = (frame->site_address % 4) * 64 + frame->encoder_address;
	/* SQC */
	block[2] = frame->sequence_counter;
	/* MFL */
	block[3] = frame->field_length;
	/* MSG: render each of the elements to the block */
	if (frame->field_length > 0)
	{
		GList *l = frame->field;
		unsigned char *ptr = &block[4];
		while (l)
		{
			UecpElement *element = l->data;
			int length = uecp_element_get_length (element);

			uecp_element_render_to_block (element, ptr);
			ptr += length;
			l = g_list_next (l);
		}
		crc = ccitt_crc (block, crc_length);
	}
	block[block_length - 2] = crc / 256;
	block[block_length - 1] = crc % 256;

	stuffed_length = uecp_byte_stuff_get_length (block, block_length);
	stuffed = (unsigned char *) g_malloc (1 + stuffed_length + 1);
	stuffed[0] = 0xfe; /* STA */
	stuffed[stuffed_length + 1] = 0xff; /* STP */
	uecp_byte_stuff_to_block (block, block_length, &stuffed[1]);
	g_free (block);
	return stuffed;
}

/* add an element to this frame if there is room left for it
 * returns FALSE in case of error (ie, too long)
 * the frame takes ownership of the element, so the calling program
 * is not responsible for freeing it anymore, nor allowed to access it
 * after adding it to the frame
 */
gboolean
uecp_frame_add_element (UecpFrame *frame, UecpElement *element)
{
	int length = uecp_element_get_length (element);

	if (frame->field_length + length >= UECP_MAX_FIELD_LENGTH)
		return FALSE;
	frame->field_length += length;
	frame->field = g_list_append (frame->field, element);
	return TRUE;
}


void
uecp_frame_set_sequence_counter (UecpFrame *frame, unsigned char counter)
{
	/*
	 * < 1, or 0, means "not used", not necessarily an error
	if (counter < 1)
	{
		g_print ("uecp: error: counter < 1\n");
		return;
	}
	*/
	frame->sequence_counter = counter;
}

void
uecp_frame_set_site_address (UecpFrame *frame, unsigned int site_address)
{
	int limit = 1 << 10;
	if (site_address >= limit)
	{
		g_print ("uecp: error: site_address more than %d\n", limit);
		return;
	}
	frame->site_address = site_address;
}

void
uecp_frame_set_encoder_address (UecpFrame *frame, unsigned int encoder_address)
{
	int limit = 1 << 6;
	if (encoder_address >= limit)
	{
		g_print ("uecp: error: encoder_address more than %d\n", limit);
		return;
	}
	frame->encoder_address = encoder_address;
}

/*
 * sets a field on the uecp frame and calculate length from it */
void
uecp_frame_set_field (UecpFrame *frame, GList *field)
{
	if (frame->field)
	{
		g_print ("uecp: error: frame->field already set\n");
		return;
	}
	frame->field = field;
	/* FIXME
	frame->field_length = uecp_field_get_length (field);
	*/
}

void
uecp_frame_debug (UecpFrame *frame)
{
	g_print ("uecp: debug: frame\n");
}

