/* messager.c - render specific UECP messages to a buffer
 *
 * Copyright (C) Thomas Vander Stichele
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <glib.h>
#include "element.h"
#include "frame.h"
#include "messager.h"

/* private functions */
/* compose a message and render it to an output stream.
 * the passed message should be allocated and is taken ownership of
 * returns NULL in case of error
 */
static guchar *
uecp_messager_compose (UecpMessager *messager, UecpMessageElementCode mec,
	               gboolean use_dsn, gboolean use_psn, int mel,
		       int medl, guchar *med)
{
	guchar *result;
	UecpFrame *frame;
	UecpElement *element;
	gint dsn = -1;
	gint psn = -1;

	if (use_dsn)
		dsn = messager->data_set_number;
	if (use_psn)
		psn = messager->program_service_number;

	/* create element */
	element = uecp_element_new ();
	if (!uecp_element_set (element, mec, dsn, psn, mel, medl, med))
	{
		uecp_element_dispose (element);
		return NULL;
	}

	/* create frame */
	frame = uecp_frame_new ();

	/* set address */
	uecp_frame_set_site_address (frame, messager->site_address);
	uecp_frame_set_encoder_address (frame, messager->encoder_address);
	uecp_frame_set_sequence_counter (frame, messager->sequence_counter);

	/* attach element */
	if (!uecp_frame_add_element (frame, element))
	{
		uecp_element_dispose (element);
		uecp_frame_dispose (frame);
		return NULL;
	}

	result = uecp_frame_render (frame);

	/* clean up */
	if (messager->sequence_counter == 255)
		messager->sequence_counter = 0; /* wrap from 255 -> 1 */
	messager->sequence_counter++;
	uecp_frame_dispose (frame);

	return result;
}

/*
 * public functions
 */

/* object-y functions */

/* create a new messager and initialize it, and return it */
UecpMessager *
uecp_messager_new ()
{
	UecpMessager *messager;

	messager = g_malloc (sizeof (UecpMessager));
	uecp_messager_reset (messager);

	return messager;
}

/* reset a messager to a zero status */
void
uecp_messager_reset (UecpMessager *messager)
{
	messager->site_address = 0;
	messager->encoder_address = 0;
	messager->data_set_number = 0;
	messager->program_service_number = 0;
	messager->sequence_counter = 1;
}

/* dispose of all resources held by messager */
void
uecp_messager_dispose (UecpMessager *messager)
{
	uecp_messager_reset (messager);
	g_free (messager);
}

/* actual message functions */

/* send the null-terminated ps string as the PS code
 * PS is always 8 chars, so stuff it with spaces if need be, and don't send
 * a MEL */
guchar *
uecp_messager_ps (UecpMessager *messager, const guchar *ps)
{
	guchar *string;
	unsigned int length;
	int i;

	/* allocate a string for PS and space-fill it */
	string = (guchar *) g_malloc (UECP_PS_SIZE);
	for (i = 0; i < UECP_PS_SIZE; ++i) *(string + i) = ' ';

	/* copy the copyable part of the passed ps */
	length = (strlen (ps) > UECP_PS_SIZE ? UECP_PS_SIZE : strlen (ps));
	memcpy (string, ps, length);
	if (length < strlen (ps))
		g_print ("warning: length of ps %d, truncated.\n",
			 strlen (ps));

	return uecp_messager_compose (messager, UECP_MEC_PS, TRUE, TRUE,
				      -1, UECP_PS_SIZE, string);
}

/* send TA and TP */
guchar *
uecp_messager_ta_tp (UecpMessager *messager, gboolean ta, gboolean tp)
{
	guchar *string;
	unsigned int length;

	string = (guchar *) g_malloc (sizeof (1));
	*string = 0;
	if (ta) *string = 1;	/* first bit */
	if (tp) *string += 2;	/* second bit */

	return uecp_messager_compose (messager, UECP_MEC_TA_TP, TRUE, TRUE,
				      -1, 1, string);
}

/* send RT;
 * rt: text to transmit (max. 64 chars)
 * if add is TRUE then the buffer is added to the previous, otherwise
 *    radiotext is flushed
 * number -> number of transmission (1 .. 15, 0 for infinite)
 * toggle -> whether or not to toggle A/B flag
 */
guchar *
uecp_messager_rt (UecpMessager *messager, const guchar *rt,
		  gboolean add, guchar number, gboolean toggle)
{
	guchar *string;
	unsigned int length; /* length of radiotext */
	guchar configuration;

	length = (strlen (rt) > UECP_RT_SIZE ? UECP_RT_SIZE : strlen (rt));
	g_print ("DEBUG: length of %s: %d\n", rt, length);
	if (length < strlen (rt))
		g_print ("warning: length of radiotext %d, truncated.\n",
			 strlen (rt));
	string = (guchar *) g_malloc (1 + length);

	/* set buffer configuration byte */
	/* bit 0 is A/B toggle flag */
	configuration = (toggle ? 1 : 0);
	/* bits 1-4 contain number of retransmissions */
	if (number > 15)
	{
		g_print ("warning: number of retransmissions > 15, set to 0\n");
		number = 0;
	}
	configuration += number << 1;
	/* bit 5 is always set to 0; 1 is reserved */
	/* bit 6 is set to 1 if message should be added */
	if (add) configuration += 1 << 6;
	/* bit 7 is reserved */
	*string = configuration;
	memcpy (string + 1, rt, length);

	return uecp_messager_compose (messager, UECP_MEC_RT, TRUE, TRUE,
				      length + 1, length + 1, string);
}
